# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Twisted specific stats utilities."""

from greplin import scales

from twisted.internet import reactor


def installStatsLoop(statsFile, statsDelay):
  """Installs an interval loop that dumps stats to a file."""

  def dumpStats():
    """Actual stats dump function."""
    scales.dumpStatsTo(statsFile)
    reactor.callLater(statsDelay, dumpStats)

  def startStats():
    """Starts the stats dump in "statsDelay" seconds."""
    reactor.callLater(statsDelay, dumpStats)

  reactor.callWhenRunning(startStats)



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for stat formatting."""

from greplin import scales
from greplin.scales import formats

import six
import unittest

try:
  import simplejson as json
except ImportError:
  import json



class Root(object):
  """Root level test class."""

  def __init__(self):
    scales.init(self)


  def getChild(self, name, collapsed):
    """Creates a child."""
    return Child(name, collapsed)



class Child(object):
  """Child test class."""

  countStat = scales.IntStat('count')


  def __init__(self, name, collapsed):
    scales.initChild(self, name).setCollapsed(collapsed)



class StatsTest(unittest.TestCase):
  """Test cases for stats classes."""

  def setUp(self):
    """Reset global state."""
    scales.reset()


  def testJsonCollapse(self):
    """Tests for collapsed child stats."""
    r = Root()
    r.getChild('here', False).countStat += 1
    r.getChild('not', True).countStat += 100

    out = six.StringIO()
    formats.jsonFormat(out)

    self.assertEquals('{"here": {"count": 1}}\n', out.getvalue())



class UnicodeFormatTest(unittest.TestCase):
  """Test cases for Unicode stat formatting."""

  UNICODE_VALUE = six.u('\u842c\u77e5\u5802')


  def testHtmlFormat(self):
    """Test generating HTML with Unicode values."""
    out = six.StringIO()
    formats.htmlFormat(out, statDict={'name': self.UNICODE_VALUE})
    result = out.getvalue()
    if six.PY2:
        value = self.UNICODE_VALUE.encode('utf8')
    else:
        value = self.UNICODE_VALUE
    self.assertTrue(value in result)


  def testJsonFormat(self):
    """Test generating JSON with Unicode values."""
    out = six.StringIO()
    stats = {'name': self.UNICODE_VALUE}
    formats.jsonFormat(out, statDict=stats)
    self.assertEquals(stats, json.loads(out.getvalue()))


  def testJsonFormatBinaryGarbage(self):
    """Make sure that JSON formatting of binary junk does not crash."""
    out = six.StringIO()
    stats = {'garbage': '\xc2\xc2 ROAR!! \0\0'}
    formats.jsonFormat(out, statDict=stats)
    self.assertEquals(json.loads(out.getvalue()), {'garbage': six.u('\xc2\xc2 ROAR!! \0\0')})





# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for pushing stat values to graphite."""

from greplin import scales
from greplin.scales import util

import os
import threading
import logging
import time
from fnmatch import fnmatch
from socket import gethostname

import six

class GraphitePusher(object):
  """A class that pushes all stat values to Graphite on-demand."""

  def __init__(self, host, port, prefix=None):
    """If prefix is given, it will be prepended to all Graphite
    stats. If it is not given, then a prefix will be derived from the
    hostname."""
    self.rules = []
    self.pruneRules = []

    self.prefix = prefix or gethostname().lower()

    if self.prefix and self.prefix[-1] != '.':
      self.prefix += '.'

    self.graphite = util.GraphiteReporter(host, port)


  def _sanitize(self, name):
    """Sanitize a name for graphite."""
    return name.strip().replace(' ', '-').replace('.', '-').replace('/', '_')


  def _forbidden(self, path, value):
    """Is a stat forbidden? Goes through the rules to find one that
    applies. Chronologically newer rules are higher-precedence than
    older ones. If no rule applies, the stat is forbidden by default."""
    if path[0] == '/':
      path = path[1:]
    for rule in reversed(self.rules):
      if isinstance(rule[1], six.string_types):
        if fnmatch(path, rule[1]):
          return not rule[0]
      elif rule[1](path, value):
        return not rule[0]
    return True # do not log by default


  def _pruned(self, path):
    """Is a stat tree node pruned?  Goes through the list of prune rules
    to find one that applies.  Chronologically newer rules are
    higher-precedence than older ones. If no rule applies, the stat is
    not pruned by default."""
    if path[0] == '/':
      path = path[1:]
    for rule in reversed(self.pruneRules):
      if isinstance(rule, six.string_types):
        if fnmatch(path, rule):
          return True
      elif rule(path):
        return True
    return False # Do not prune by default


  def push(self, statsDict=None, prefix=None, path=None):
    """Push stat values out to Graphite."""
    if statsDict is None:
      statsDict = scales.getStats()
    prefix = prefix or self.prefix
    path = path or '/'

    for name, value in list(statsDict.items()):
      name = str(name)
      subpath = os.path.join(path, name)

      if self._pruned(subpath):
        continue

      if hasattr(value, '__call__'):
        try:
          value = value()
        except:                       # pylint: disable=W0702
          value = None
          logging.exception('Error when calling stat function for graphite push')

      if hasattr(value, 'iteritems'):
        self.push(value, '%s%s.' % (prefix, self._sanitize(name)), subpath)
      elif self._forbidden(subpath, value):
        continue

      if six.PY3:
        type_values = (int, float)
      else:
        type_values = (int, long, float)

      if type(value) in type_values and len(name) < 500:
        self.graphite.log(prefix + self._sanitize(name), value)


  def _addRule(self, isWhitelist, rule):
    """Add an (isWhitelist, rule) pair to the rule list."""
    if isinstance(rule, six.string_types) or hasattr(rule, '__call__'):
      self.rules.append((isWhitelist, rule))
    else:
      raise TypeError('Graphite logging rules must be glob pattern or callable. Invalid: %r' % rule)


  def allow(self, rule):
    """Append a whitelisting rule to the chain. The rule is either a function (called
    with the stat name and its value, returns True if it matches), or a Bash-style
    wildcard pattern, such as 'foo.*.bar'."""
    self._addRule(True, rule)


  def forbid(self, rule):
    """Append a blacklisting rule to the chain. The rule is either a function (called
    with the stat name and its value, returns True if it matches), or a Bash-style
    wildcard pattern, such as 'foo.*.bar'."""
    self._addRule(False, rule)


  def prune(self, rule):
    """Append a rule that stops traversal at a branch node."""
    self.pruneRules.append(rule)



class GraphitePeriodicPusher(threading.Thread, GraphitePusher):
  """A thread that periodically pushes all stat values to Graphite."""

  def __init__(self, host, port, prefix=None, period=60):
    """If prefix is given, it will be prepended to all Graphite
    stats. If it is not given, then a prefix will be derived from the
    hostname."""
    GraphitePusher.__init__(self, host, port, prefix)
    threading.Thread.__init__(self)
    self.daemon = True

    self.period = period


  def run(self):
    """Loop forever, pushing out stats."""
    self.graphite.start()
    while True:
      logging.info('Graphite pusher is sleeping for %d seconds', self.period)
      time.sleep(self.period)
      logging.info('Pushing stats to Graphite')
      try:
        self.push()
        logging.info('Done pushing stats to Graphite')
      except:
        logging.exception('Exception while pushing stats to Graphite')
        raise



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the stats module."""

from greplin import scales

import unittest



class Root1(object):
  """Root level test class."""

  stateStat = scales.Stat('state')
  errorsStat = scales.IntDictStat('errors')
  activeUrlsStat = scales.IntDictStat('activeUrls', autoDelete=True)


  def __init__(self):
    scales.init(self, 'path/to/A')


  def getChild(self, cls, *args):
    """Creates a child."""
    return cls(*args)



class Root2(object):
  """Root level test class."""

  def __init__(self):
    scales.init(self, 'B')


  def getChild(self, cls):
    """Creates a child."""
    return cls()



class AggregatingRoot(object):
  """Root level test class with aggregation."""

  countStat = scales.SumAggregationStat('count')
  stateStat = scales.HistogramAggregationStat('state')
  errorsStat = scales.IntDictSumAggregationStat('errors')


  def __init__(self):
    scales.init(self, 'Root')


  def getChild(self, cls, *args):
    """Creates a child."""
    return cls(*args)



class AggregatingRootSubclass(AggregatingRoot):
  """Subclass of a class with aggregates."""



class TypedChild(object):
  """Child level test class."""

  countStat = scales.IntStat('count')


  def __init__(self):
    scales.initChildOfType(self, 'C')



class Child(object):
  """Child level test class."""

  countStat = scales.IntStat('count')
  stateStat = scales.Stat('state')
  errorsStat = scales.IntDictStat('errors')


  def __init__(self, name='C'):
    scales.initChild(self, name)



class DynamicRoot(object):
  """Root class with a dynamic stat."""

  value = 100
  dynamicStat = scales.Stat('dynamic')


  def __init__(self):
    scales.init(self)
    self.dynamicStat = lambda: DynamicRoot.value



class StatsTest(unittest.TestCase):
  """Test cases for stats classes."""


  def setUp(self):
    """Reset global state."""
    scales.reset()


  def testChildTypeStats(self):
    """Tests for child stats with typed children (auto-numbered)."""
    a = Root1()
    a.stateStat = 'abc'
    c = a.getChild(TypedChild)
    c.countStat += 1
    b = Root2()
    c = b.getChild(TypedChild)
    c.countStat += 2

    self.assertEquals({
      'path': {
        'to': {
          'A': {
            'state': 'abc',
            'C': {
              '1': {'count': 1}
            }
          }
        }
      },
      'B': {
        'C': {
          '2': {'count': 2}
        },
      }
    }, scales.getStats())


  def testChildStats(self):
    """Tests for child scales."""
    a = Root1()
    a.stateStat = 'abc'
    c = a.getChild(Child)
    c.countStat += 1
    b = Root2()
    c = b.getChild(Child)
    c.countStat += 2

    self.assertEquals({
      'path': {
        'to': {
          'A': {
            'state': 'abc',
            'C': {
              'count': 1
            }
          }
        }
      },
      'B': {
        'C': {
          'count': 2
        },
      }
    }, scales.getStats())


  def testMultilevelChild(self):
    """Tests for multi-level child stats."""
    a = Root1()
    c = a.getChild(Child, 'sub/path')
    c.countStat += 1

    self.assertEquals({
      'path': {
        'to': {
          'A': {
            'sub': {
              'path': {
                'count': 1
              }
            }
          }
        }
      }
    }, scales.getStats())


  def testStatSum(self):
    """Tests for summed stats."""
    self.helpTestStatSum(AggregatingRoot())


  def testStatSumWithSubclassRoot(self):
    """Tests for summed stats."""
    self.helpTestStatSum(AggregatingRootSubclass())


  def helpTestStatSum(self, a):
    """Helps test summed stats."""
    c = a.getChild(Child)

    self.assertEquals({
      'Root': {
        'C': {},
      }
    }, scales.getStats())

    c.countStat += 2

    self.assertEquals({
      'Root': {
        'count': 2,
        'C': {
          'count': 2
        },
      }
    }, scales.getStats())

    d = a.getChild(Child, 'D')
    self.assertEquals({
      'Root': {
        'count': 2,
        'C': {
          'count': 2
        },
        'D': {}
      }
    }, scales.getStats())

    c.countStat -= 1
    d.countStat += 5
    self.assertEquals({
      'Root': {
        'count': 6,
        'C': {
          'count': 1
        },
        'D': {
          'count': 5
        }
      }
    }, scales.getStats())


  def testStatHistogram(self):
    """Tests for stats aggregated in to a histogram."""
    a = AggregatingRoot()
    c = a.getChild(Child)
    d = a.getChild(Child, 'D')

    # Do it twice to make sure its idempotent.
    for _ in range(2):
      c.stateStat = 'good'
      d.stateStat = 'bad'
      self.assertEquals({
        'Root': {
          'state': {
            'good': 1,
            'bad': 1
          },
          'C': {
            'state': 'good'
          },
          'D': {
            'state': 'bad'
          }
        }
      }, scales.getStats())

    c.stateStat = 'great'
    d.stateStat = 'great'
    self.assertEquals({
      'Root': {
        'state': {
          'great': 2,
          'good': 0,
          'bad': 0
        },
        'C': {
          'state': 'great'
        },
        'D': {
          'state': 'great'
        }
      }
    }, scales.getStats())



  def testIntDictStats(self):
    """Tests for int dict stats."""
    a = Root1()
    a.errorsStat['400'] += 1
    a.errorsStat['400'] += 2
    a.errorsStat['404'] += 100
    a.errorsStat['400'] -= 3

    a.activeUrlsStat['http://www.greplin.com'] += 1
    a.activeUrlsStat['http://www.google.com'] += 2
    a.activeUrlsStat['http://www.greplin.com'] -= 1

    self.assertEquals({
      'path': {
        'to': {
          'A': {
            'errors': {
              '400': 0,
              '404': 100
            },
            'activeUrls': {
              'http://www.google.com': 2
            }
          }
        }
      }
    }, scales.getStats())


  def testIntDictStatsAggregation(self):
    """Tests for int dict stats."""
    root = AggregatingRoot()

    errorHolder = root.getChild(Child)

    errorHolder.errorsStat['400'] += 1
    errorHolder.errorsStat['400'] += 2
    errorHolder.errorsStat['404'] += 100
    errorHolder.errorsStat['400'] += 1

    self.assertEquals({
      'Root': {
        'errors': {
          '400': 4,
          '404': 100
        },
        'C': {
          'errors': {
            '400': 4,
            '404': 100
          }
        }
      }
    }, scales.getStats())


  def testDynamic(self):
    """Tests for dynamic stats."""
    DynamicRoot()
    self.assertEquals(100, scales.getStats()['dynamic']())

    DynamicRoot.value = 200
    self.assertEquals(200, scales.getStats()['dynamic']())


  def testCollection(self):
    """Tests for a stat collection."""
    collection = scales.collection('/thePath', scales.IntStat('count'), scales.IntDictStat('histo'))
    collection.count += 100
    collection.histo['cheese'] += 12300
    collection.histo['cheese'] += 45

    self.assertEquals({
      'thePath': {
        'count': 100,
        'histo': {
          'cheese': 12345
        }
      }
    }, scales.getStats())



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stat aggregation tests."""

import re

from greplin.scales import aggregation

import unittest



class AggregationTest(unittest.TestCase):
  """Test cases for stat aggregation classes."""

  def testNoData(self):
    "This used to infinite loop."
    agg = aggregation.Aggregation({
      'a': {
        '*': [aggregation.Sum()]
      }
    })
    agg.addSource('source1', {'a': {}})
    agg.result()


  def testRegex(self):
    "Test regexes in aggregation keys"
    agg = aggregation.Aggregation({
        'a' : {
            ('success', re.compile("[1-3][0-9][0-9]")):  [aggregation.Sum(dataFormat = aggregation.DataFormats.DIRECT)],
            ('error', re.compile("[4-5][0-9][0-9]")):  [aggregation.Sum(dataFormat = aggregation.DataFormats.DIRECT)]
        }})
    agg.addSource('source1', {'a': {'200': 10, '302': 10, '404': 1, '500': 3}})
    result = agg.result()
    self.assertEquals(result['a']['success']['sum'], 20)
    self.assertEquals(result['a']['error']['sum'], 4)



if __name__ == '__main__':
  unittest.main()



# Copyright 2012 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the util module."""

from greplin.scales import util

import unittest



class AtomicValueTest(unittest.TestCase):
  """Tests for atomic values."""

  def testUpdate(self):
    """Test update functions."""
    v = util.AtomicValue('hello, world')
    self.assertEqual(v.update(len), ('hello, world', len('hello, world')))
    self.assertEqual(v.value, len('hello, world'))


  def testGetAndSet(self):
    """Test get-and-set."""
    v = util.AtomicValue(42)
    self.assertEqual(v.getAndSet(666), 42)
    self.assertEqual(v.value, 666)


  def testAddAndGet(self):
    """Test add-and-get."""
    v = util.AtomicValue(42)
    self.assertEqual(v.addAndGet(8), 50)
    self.assertEqual(v.value, 50)



import time

class BasicClock(object):
  """ Abstraction between things that use sources of time and the rest of the system
      this allows for independent clocks to be integrated with potentially different
      levels of granularity """

  def time(self):
    return time.time()

def getClock():
    """ Returns the best, most accurate clock possible """
    return BasicClock()



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for metering values"""

try:
  from UserDict import UserDict
except ImportError:
  from collections import UserDict
from greplin.scales import Stat
from greplin.scales.timer import RepeatTimer
from greplin.scales.util import EWMA

TICKERS = []
TICKER_THREAD = RepeatTimer(5, lambda: [t() for t in TICKERS])



class MeterStatDict(UserDict):
  """Stores the meters for MeterStat. Expects to be ticked every 5 seconds."""

  def __init__(self):
    UserDict.__init__(self)
    self._m1 = EWMA.oneMinute()
    self._m5 = EWMA.fiveMinute()
    self._m15 = EWMA.fifteenMinute()
    self._meters = (self._m1, self._m5, self._m15)
    TICKERS.append(self.tick)

    self['unit'] = 'per second'
    self['count'] = 0


  def __getitem__(self, item):
    if item in self:
      return UserDict.__getitem__(self, item)
    else:
      return 0.0


  def tick(self):
    """Updates meters"""
    for m in self._meters:
      m.tick()
    self['m1'] = self._m1.rate
    self['m5'] = self._m5.rate
    self['m15'] = self._m15.rate


  def mark(self, value=1):
    """Updates the dictionary."""

    self['count'] += value
    for m in self._meters:
      m.update(value)



class MeterStat(Stat):
  """A stat that stores m1, m5, m15. Updated every 5 seconds via TICKER_THREAD."""

  def __init__(self, name, _=None):
    Stat.__init__(self, name, None)


  def _getDefault(self, _):
    """Returns a default MeterStatDict"""
    return MeterStatDict()


  def __set__(self, instance, value):
    self.__get__(instance, None).mark(value)



class MeterDict(UserDict):
  """Dictionary of meters."""

  def __init__(self, parent, instance):
    UserDict.__init__(self)
    self.parent = parent
    self.instance = instance


  def __getitem__(self, item):
    if item in self:
      return UserDict.__getitem__(self, item)
    else:
      meter = MeterStatDict()
      self[item] = meter
      return meter



class MeterDictStat(Stat):
  """Dictionary stat value class."""

  def _getDefault(self, instance):
    return MeterDict(self, instance)



# Copyright (c) 2009 Geoffrey Foster. Portions by the Scales Authors.
# pylint: disable=W9921, C0103
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.


"""
A Timer implementation that repeats every interval. Vaguely based on

http://g-off.net/software/a-python-repeatable-threadingtimer-class

Modified to remove the Event signal as we never intend to cancel it.
This was done primarily for compatibility with libraries like gevent.
"""

from six.moves._thread import start_new_thread
from time import sleep


def RepeatTimer(interval, function, iterations=0, *args, **kwargs):
  """Repeating timer. Returns a thread id."""

  def __repeat_timer(interval, function, iterations, args, kwargs):
    """Inner function, run in background thread."""
    count = 0
    while iterations <= 0 or count < iterations:
      sleep(interval)
      function(*args, **kwargs)
      count += 1

  return start_new_thread(__repeat_timer, (interval, function, iterations, args, kwargs))



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for tracking system statistics."""

import collections
import inspect
import itertools
import gc
import six
import unittest
try:
  # Prefer simplejson for speed.
  import simplejson as json
except ImportError:
  import json
import time
from contextlib import contextmanager

try:
  from UserDict import UserDict
except ImportError:
  from collections import UserDict
from greplin.scales.samplestats import ExponentiallyDecayingReservoir

ID_KEY = '__STATS__id'


NEXT_ID = itertools.count()


def statsId(obj):
  """Gets a unique ID for each object."""
  if hasattr(obj, ID_KEY):
    return getattr(obj, ID_KEY)
  newId = next(NEXT_ID)
  setattr(obj, ID_KEY, newId)
  return newId


def init(obj, context=None):
  """Initializes stats collection in the given object at the given context path.  Defaults to root-level stats."""
  return _Stats.init(obj, context)


def initChild(obj, name):
  """Initializes stats collection in the given object as a child of the object that created it."""
  return _Stats.initChild(obj, name, '')


def initChildOfType(obj, name, subContext=None):
  """Initializes stats collection in the given object as a child of the object that created it."""
  return _Stats.initChild(obj, name, subContext)


def reset():
  """Resets stats state - should only be called from tests."""
  _Stats.reset()


def getStats():
  """Gets the stats dict."""
  return _Stats.stats


def setCollapsed(path):
  """Sets a stat as collapsed."""
  return _Stats.setCollapsed(path)



class StatContainer(UserDict):
  """Container of stats.  Also contains configuration of how the container should be displayed."""

  def __init__(self):
    UserDict.__init__(self)
    self.__collapsed = False


  def setCollapsed(self, isCollapsed):
    """Sets whether the stat container displays as collapsed by default."""
    self.__collapsed = isCollapsed


  def isCollapsed(self):
    """Returns whether the stat container displays as collapsed by default."""
    return self.__collapsed



class _Stats(object):
  """Static class for stats aggregation."""

  stats = StatContainer()

  parentMap = {}

  containerMap = {}

  subId = 0


  @classmethod
  def reset(cls):
    """Resets the static state.  Should only be called by tests."""
    cls.stats = StatContainer()
    cls.parentMap = {}
    cls.containerMap = {}
    cls.subId = 0
    for stat in gc.get_objects():
      if isinstance(stat, Stat):
        stat._aggregators = {}


  @classmethod
  def init(cls, obj, context):
    """Implementation of init."""
    addr = statsId(obj)
    if addr not in cls.containerMap:
      cls.containerMap[addr] = cls.__getStatContainer(context)
    return cls.containerMap[addr]


  @classmethod
  def __getSelf(cls, frame):
    """Extracts the self object out of a stack frame."""
    return inspect.getargvalues(frame).locals.get('self', None)


  @classmethod
  def initChild(cls, obj, name, subContext, parent = None):
    """Implementation of initChild."""
    addr = statsId(obj)
    if addr not in cls.containerMap:
      if not parent:
        # Find out the parent of the calling object by going back through the call stack until a self != this.
        f = inspect.currentframe()
        while not cls.__getSelf(f):
          f = f.f_back
        this = cls.__getSelf(f)
        f = f.f_back
        while cls.__getSelf(f) == this or not cls.__getSelf(f):
          f = f.f_back
        parent = cls.__getSelf(f)

      # Default subcontext to an autoincrementing ID.
      if subContext is None:
        cls.subId += 1
        subContext = cls.subId

      if subContext is not '':
        path = '%s/%s' % (name, subContext)
      else:
        path = name

      # Now that we have the name, create an entry for this object.
      cls.parentMap[addr] = parent
      container = cls.getContainerForObject(statsId(parent))
      if not container and isinstance(parent, unittest.TestCase):
        cls.init(parent, '/test-case')
      cls.containerMap[addr] = cls.__getStatContainer(path, cls.getContainerForObject(statsId(parent)))
    return cls.containerMap[addr]


  @classmethod
  def __getStatContainer(cls, context, parent=None):
    """Get the stat container for the given context under the given parent."""
    container = parent
    if container is None:
      container = cls.stats
    if context is not None:
      context = str(context).lstrip('/')
      for key in context.split('/'):
        container.setdefault(key, StatContainer())
        container = container[key]
    return container


  @classmethod
  def getContainerForObject(cls, instanceId):
    """Get the stat container for the given object."""
    return cls.containerMap.get(instanceId, None)


  @classmethod
  def getStat(cls, obj, name):
    """Gets the stat for the given object with the given name, or None if no such stat exists."""
    objClass = type(obj)
    for theClass in objClass.__mro__:
      if theClass == object:
        break
      for value in theClass.__dict__.values():
        if isinstance(value, Stat) and value.getName() == name:
          return value


  @classmethod
  def getAggregator(cls, instanceId, name):
    """Gets the aggregate stat for the given stat."""
    parent = cls.parentMap.get(instanceId)
    while parent:
      stat = cls.getStat(parent, name)
      if stat:
        return stat, parent
      parent = cls.parentMap.get(statsId(parent))


  @classmethod
  def setCollapsed(cls, path):
    """Collapses a stat."""
    cls.__getStatContainer(path).setCollapsed(True)



class Stat(object):
  """Basic stat value class."""

  def __init__(self, name, value='', logger = None):
    self.__name = name
    self.__default = value
    self._logger = logger
    self._aggregators = {}


  def getName(self):
    """Gets the name of the stat."""
    return self.__name


  def __get__(self, instance, _):
    container = _Stats.getContainerForObject(statsId(instance))
    if self.__name not in container:
      container[self.__name] = self._getDefault(instance)
    return container[self.__name]


  def _getInit(self):
    """Internal method to return the initial value for a stat that is never set."""
    return self.__default


  def _getDefault(self, _):
    """Internal method to return the default for a stat that hasn't stored a value yet."""
    return self.__default


  def _aggregate(self, instanceId, container, value, subKey = None):
    """Performs stat aggregation."""

    # Get the aggregator.
    if instanceId not in self._aggregators:
      self._aggregators[instanceId] = _Stats.getAggregator(instanceId, self.__name)
    aggregator = self._aggregators[instanceId]

    # If we are aggregating, get the old value.
    if aggregator:
      oldValue = container.get(self.__name)
      if subKey:
        oldValue = oldValue[subKey]
        aggregator[0].update(aggregator[1], oldValue, value, subKey)
      else:
        aggregator[0].update(aggregator[1], oldValue, value)


  def __set__(self, instance, value):
    instanceId = statsId(instance)

    container = _Stats.getContainerForObject(instanceId)
    self._aggregate(instanceId, container, value)
    container[self.__name] = value

    if self._logger:
      self._logger('Updated stat "%s" with value: %s' % (self.__name, value))


  def updateItem(self, instance, subKey, value):
    """Updates a child value.  Must be called before the update has actually occurred."""
    instanceId = statsId(instance)

    container = _Stats.getContainerForObject(instanceId)
    self._aggregate(instanceId, container, value, subKey)


  def logger(self, logger):
    """Log textual updates about this value to the given function."""
    self._logger = logger
    return self



class IntStat(Stat):
  """Integer stat value class."""

  def __init__(self, name, value=0):
    Stat.__init__(self, name, value)



class DoubleStat(Stat):
  """Double stat value class."""

  def __init__(self, name, value=0.0):
    Stat.__init__(self, name, value)



class IntDict(UserDict):
  """Dictionary of integers."""

  def __init__(self, parent, instance, autoDelete=False):
    UserDict.__init__(self)
    self.parent = parent
    self.instance = instance
    self.autoDelete = autoDelete


  def __getitem__(self, item):
    if item in self:
      return UserDict.__getitem__(self, item)
    else:
      return 0


  def __setitem__(self, key, value):
    self.parent.updateItem(self.instance, key, value)
    if value or not self.autoDelete:
      UserDict.__setitem__(self, key, value)
    elif UserDict.__contains__(self, key):
      UserDict.__delitem__(self, key)



class IntDictStat(Stat):
  """Dictionary stat value class."""

  def __init__(self, name, autoDelete = False):
    Stat.__init__(self, name)
    self.autoDelete = autoDelete


  def _getDefault(self, instance):
    return IntDict(self, instance, self.autoDelete)



class StringDict(UserDict):
  """Dictionary of strings."""

  def __init__(self, parent, instance):
    UserDict.__init__(self)
    self.parent = parent
    self.instance = instance


  def __getitem__(self, item):
    if item in self:
      return UserDict.__getitem__(self, item)
    else:
      return ''


  def __setitem__(self, key, value):
    self.parent.updateItem(self.instance, key, value)
    UserDict.__setitem__(self, key, value)



class StringDictStat(Stat):
  """Dictionary stat value class."""

  def _getDefault(self, instance):
    return StringDict(self, instance)



class AggregationStat(Stat):
  """A stat that aggregates child stats."""

  def __init__(self, name, value):
    Stat.__init__(self, name, value)


  def update(self, instance, oldValue, newValue):
    """Updates the aggregate based on a change in the child value."""
    raise NotImplementedError



class ChildAggregationStat(Stat):
  """A stat that aggregates values within child stats."""

  def __init__(self, name, value):
    Stat.__init__(self, name, value)


  def update(self, instance, oldValue, newValue, subKey):
    """Updates the aggregate based on a change in the child sub-value."""



class SumAggregationStat(AggregationStat):
  """A stat that aggregates child stats in to a sum."""

  def __init__(self, name):
    AggregationStat.__init__(self, name, 0)


  def update(self, instance, oldValue, newValue):
    """Updates the aggregate based on a change in the child value."""
    self.__set__(instance,
                 self.__get__(instance, None) + newValue - (oldValue or 0))



class HistogramAggregationStat(AggregationStat):
  """A stat that aggregates child stats in to a histogram counting each unique child value."""

  def __init__(self, name, autoDelete = False):
    AggregationStat.__init__(self, name, None)
    self.autoDelete = autoDelete


  def _getDefault(self, _):
    return collections.defaultdict(int)


  def update(self, instance, oldValue, newValue):
    """Updates the aggregate based on a change in the child value."""
    histogram = self.__get__(instance, None)
    if oldValue:
      histogram[oldValue] -= 1
      if self.autoDelete and histogram[oldValue] == 0:
        del histogram[oldValue]
    if newValue:
      histogram[newValue] += 1



class IntDictSumAggregationStat(ChildAggregationStat):
  """A stat that aggregates child int dict stats in to a int dict summing child values."""

  def __init__(self, name):
    ChildAggregationStat.__init__(self, name, None)


  def _getDefault(self, _):
    return collections.defaultdict(int)


  def update(self, instance, oldValue, newValue, subKey):
    """Updates the aggregate based on a change in the child value."""
    histogram = self.__get__(instance, None)
    histogram[subKey] += newValue - oldValue



class PmfStatDict(UserDict):
  """Ugly hack defaultdict-like thing."""

  class TimeManager(object):
    """Context manager for timing."""

    def __init__(self, container):
      self.container = container
      self.msg99 = None
      self.start = None
      self.__discard = False


    def __enter__(self):
      self.start = time.time()
      return self


    def __exit__(self, *_):
      if not self.__discard:
        latency = time.time() - self.start
        self.container.addValue(latency)

        if self.container.percentile99 is not None and latency >= self.container.percentile99:
          if self.msg99 is not None:
            logger, msg, args = self.msg99
            logger.warn(msg, *args)


    def warn99(self, logger, msg, *args):
      """If this time through the timed section of code takes longer
      than the 99th percentile time, then this will call
      logger.warn(msg, *args) at the end of the section."""
      self.msg99 = (logger, msg, args)


    def discard(self):
      """Discard this sample."""
      self.__discard = True


  def __init__(self, sample = None):
    UserDict.__init__(self)
    if sample:
        self.__sample = sample
    else:
        self.__sample = ExponentiallyDecayingReservoir()
    self.__timestamp = 0
    self.percentile99 = None
    self['count'] = 0


  def __getitem__(self, item):
    if item in self:
      return UserDict.__getitem__(self, item)
    else:
      return 0.0


  def addValue(self, value):
    """Updates the dictionary."""
    self['count'] += 1
    self.__sample.update(value)
    if time.time() > self.__timestamp + 20 and len(self.__sample) > 1:
      self.__timestamp = time.time()
      self['min'] = self.__sample.min
      self['max'] = self.__sample.max
      self['mean'] = self.__sample.mean
      self['stddev'] = self.__sample.stddev

      percentiles = self.__sample.percentiles([0.5, 0.75, 0.95, 0.98, 0.99, 0.999])
      self['median'] = percentiles[0]
      self['75percentile'] = percentiles[1]
      self['95percentile'] = percentiles[2]
      self['98percentile'] = percentiles[3]
      self['99percentile'] = percentiles[4]
      self.percentile99 = percentiles[4]
      self['999percentile'] = percentiles[5]


  def time(self):
    """Measure the time this section of code takes. For use in with statements."""
    return self.TimeManager(self)



class PmfStat(Stat):
  """A stat that stores min, max, mean, standard deviation, and some
  percentiles for arbitrary floating-point data. This is potentially a
  bit expensive, so its child values are only updated once every
  twenty seconds."""

  def __init__(self, name, _=None):
    Stat.__init__(self, name, None)


  def _getDefault(self, _):
    return PmfStatDict()


  def __set__(self, instance, value):
    self.__get__(instance, None).addValue(value)



class NamedPmfDict(UserDict):
  """Dictionary of strings."""

  def __init__(self):
    UserDict.__init__(self)


  def __getitem__(self, item):
    if item not in self:
      value = PmfStatDict()
      UserDict.__setitem__(self, item, value)
    return UserDict.__getitem__(self, item)


  def __setitem__(self, key, value):
    self[key].addValue(value)



class NamedPmfDictStat(Stat):
  """Dictionary stat value class.  Not compatible with aggregation at this time."""

  def _getDefault(self, _):
    return NamedPmfDict()



class StateTimeStatDict(UserDict):
  """Special dict that tracks time spent in current state."""

  def __init__(self, parent, instance):
    UserDict.__init__(self)
    self.parent = parent
    self.instance = instance


  def __getitem__(self, item):
    if item in self:
      value = UserDict.__getitem__(self, item)
    else:
      value = 0.0

    if item is not None and item == self.parent.state:
      return value + (time.time() - self.parent.time)
    else:
      return value


  def incr(self, item, value):
    """Increment a key by the given amount."""
    if item in self:
      old = UserDict.__getitem__(self, item)
    else:
      old = 0.0
    self[item] = old + value


  @contextmanager
  def acquire(self):
    """Assuming that the current state is an integer (it defaults to
    zero), increment it for the duration of the body of the with
    statement."""
    self.parent.__set__(self.instance, self.parent.state + 1)
    try:
      yield
    finally:
      self.parent.__set__(self.instance, self.parent.state - 1)



class StateTimeStat(Stat):
  """A stat that stores the amount of time spent in each of a finite
  number of discrete states. This can be used to track things like
  number of concurrent users, connection pool usage, how much time a
  finite state machine spends in each state, or anything like that. To
  use it, just set the stat to the new state every time the state
  changes."""

  def __init__(self, name, _=None):
    Stat.__init__(self, name, None)
    self.state = 0
    self.time = None


  def _getDefault(self, instance):
    return StateTimeStatDict(self, instance)


  def __set__(self, instance, value):
    if value == self.state:
      return
    histogram = self.__get__(instance, None)
    now = time.time()
    if self.time is not None:
      histogram.incr(self.state, now - self.time)
    self.state = value
    self.time = now



def filterCollapsedItems(data):
  """Return a filtered iteration over a list of items."""
  return ((key, value)\
          for key, value in six.iteritems(data) \
          if not (isinstance(value, StatContainer) and value.isCollapsed()))



class StatContainerEncoder(json.JSONEncoder):
  """JSON encoding that takes in to account collapsed stat containers and stat functions."""

  # pylint: disable=E0202
  def default(self, obj):
    if isinstance(obj, UserDict):
      return dict(filterCollapsedItems(obj.data))

    elif hasattr(obj, '__call__'):
      return obj()

    else:
      return json.JSONEncoder.default(self, obj)


def dumpStatsTo(filename):
  """Writes the stats dict to filanem"""
  with open(filename, 'w') as f:
    latest = getStats()
    latest['last-updated'] = time.time()
    json.dump(getStats(), f, cls=StatContainerEncoder)



def collection(path, *stats):
  """Creates a named stats collection object."""

  def initMethod(self):
    """Init method for the underlying stat object's class."""
    init(self, path)

  attributes = {'__init__': initMethod}
  for stat in stats:
    attributes[stat.getName()] = stat
  newClass = type('Stats:%s' % path, (object,), attributes)
  instance = newClass()
  for stat in stats:
    default = stat._getInit() # Consider this method package-protected. # pylint: disable=W0212
    if default:
      setattr(instance, stat.getName(), default)
  return instance



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample statistics. Based on the Java code in Yammer metrics."""

import random
from math import sqrt, floor, exp

from .clock import getClock

def _bounded_exp(value):
  """ Emulate the Java version of exp """
  try:
    return exp(value)
  except OverflowError:
    return float("inf")

class Sampler(object):
  """Class to do simple sampling over values"""
  def __init__(self):
    self.min = float('inf')
    self.max = float('-inf')

  def __len__(self):
    return 0

  def samples(self):
    return []

  def update(self, value):
    self.min = min(self.min, value)
    self.max = max(self.max, value)

  @property
  def mean(self):
    """Return the sample mean."""
    if len(self) == 0:
      return float('NaN')
    arr = self.samples()
    return sum(arr) / float(len(arr))


  @property
  def stddev(self):
    """Return the sample standard deviation."""
    if len(self) < 2:
      return float('NaN')
    # The stupidest algorithm, but it works fine.
    try:
      arr = self.samples()
      mean = sum(arr) / len(arr)
      bigsum = 0.0
      for x in arr:
        bigsum += (x - mean)**2
      return sqrt(bigsum / (len(arr) - 1))
    except ZeroDivisionError:
      return float('NaN')


  def percentiles(self, percentiles):
    """Given a list of percentiles (floats between 0 and 1), return a
    list of the values at those percentiles, interpolating if
    necessary."""
    try:
      scores = [0.0]*len(percentiles)

      if self.count > 0:
        values = self.samples()
        values.sort()

        for i in range(len(percentiles)):
          p = percentiles[i]
          pos = p * (len(values) + 1)
          if pos < 1:
            scores[i] = values[0]
          elif pos > len(values):
            scores[i] = values[-1]
          else:
            upper, lower = values[int(pos - 1)], values[int(pos)]
            scores[i] = lower + (pos - floor(pos)) * (upper - lower)

      return scores
    except IndexError:
      return [float('NaN')] * len(percentiles)


class ExponentiallyDecayingReservoir(Sampler):
  """
    An exponentially-decaying random reservoir of. Uses Cormode et al's
    forward-decaying priority reservoir sampling method to produce a statistically representative
    sampling reservoir, exponentially biased towards newer entries.

    `Cormode et al. Forward Decay: A Practical Time Decay Model for Streaming Systems. ICDE '09
      http://dimacs.rutgers.edu/~graham/pubs/papers/fwddecay.pdf`

    This is a straight transliteration of the Yammer metrics version from java to python, whilst
    staring gently at the Cormode paper.
  """

  DEFAULT_SIZE = 1028
  DEFAULT_ALPHA = 0.015
  DEFAULT_RESCALE_THRESHOLD = 3600

  def __init__(self, size=DEFAULT_SIZE, alpha=DEFAULT_ALPHA, rescale_threshold=DEFAULT_RESCALE_THRESHOLD, clock=getClock()):
    """
      Creates a new ExponentiallyDecayingReservoir of 1028 elements, which offers a 99.9%
      confidence level with a 5% margin of error assuming a normal distribution, and an alpha
      factor of 0.015, which heavily biases the reservoir to the past 5 minutes of measurements.

      @param size  the number of samples to keep in the sampling reservoir
      @param alpha the exponential decay factor; the higher this is, the more biased the reservoir
                will be towards newer values
      @param rescale_threshold the time period over which to decay
    """
    super(ExponentiallyDecayingReservoir, self).__init__()

    self.values = {}
    self.alpha = alpha
    self.size = size
    self.clock = clock
    self.rescale_threshold = rescale_threshold
    self.count = 0
    self.startTime = self.clock.time()
    self.nextScaleTime = self.clock.time() + self.rescale_threshold

  def __len__(self):
    return min(self.size, self.count)

  def clear(self):
    """ Clear the samples. """
    self.__init__(size=self.size, alpha=self.alpha, clock=self.clock)

  def update(self, value):
    """
      Adds an old value with a fixed timestamp to the reservoir.
      @param value     the value to be added
    """
    super(ExponentiallyDecayingReservoir, self).update(value)

    timestamp = self.clock.time()

    self.__rescaleIfNeeded()
    priority = self.__weight(timestamp - self.startTime) / random.random()

    self.count += 1
    if (self.count <= self.size):
      self.values[priority] = value
    else:
      first = min(self.values)

      if first < priority and priority not in self.values:
        self.values[priority] = value
        while first not in self.values:
          first = min(self.values)

        del self.values[first]

  def __rescaleIfNeeded(self):
    now = self.clock.time()
    nextTick = self.nextScaleTime
    if now >= nextTick:
      self.__rescale(now, nextTick)

  def __weight(self, t):
    weight = self.alpha * t
    return _bounded_exp(weight)

  def __rescale(self, now, nextValue):
    if self.nextScaleTime == nextValue:
      self.nextScaleTime = now + self.rescale_threshold
      oldStartTime = self.startTime;
      self.startTime = self.clock.time()
      keys = list(self.values.keys())
      keys.sort()
      delKeys = []

      for key in keys:
        value = self.values[key]
        delKeys.append(key)
        newKey = key * _bounded_exp(-self.alpha * (self.startTime - oldStartTime))
        self.values[newKey] = value

      for key in delKeys:
        del self.values[key]

      self.count = len(self.values)

  def samples(self):
    return list(self.values.values())[:len(self)]

class UniformSample(Sampler):
  """A uniform sample of values over time."""

  def __init__(self):
    """Create an empty sample."""
    super(UniformSample, self).__init__()

    self.sample = [0.0] * 1028
    self.count = 0

  def clear(self):
    """Clear the sample."""
    for i in range(len(self.sample)):
      self.sample[i] = 0.0
    self.count = 0

  def __len__(self):
    """Number of samples stored."""
    return min(len(self.sample), self.count)

  def update(self, value):
    """Add a value to the sample."""
    super(UniformSample, self).update(value)

    self.count += 1
    c = self.count
    if c < len(self.sample):
      self.sample[c-1] = value
    else:
      r = random.randint(0, c)
      if r < len(self.sample):
        self.sample[r] = value


  def __iter__(self):
    """Return an iterator of the values in the sample."""
    return iter(self.sample[:len(self)])

  def samples(self):
    return self.sample[:len(self)]

# vim: set et fenc=utf-8 ff=unix sts=2 sw=2 ts=2 :



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful utility functions and objects."""

from six.moves.queue import Queue
from math import exp

import logging
import random
import socket
import threading
import time


def lookup(source, keys, fallback = None):
  """Traverses the source, looking up each key.  Returns None if can't find anything instead of raising an exception."""
  try:
    for key in keys:
      source = source[key]
    return source
  except (KeyError, AttributeError, TypeError):
    return fallback



class GraphiteReporter(threading.Thread):
  """A graphite reporter thread."""

  def __init__(self, host, port, maxQueueSize=10000):
    """Connect to a Graphite server on host:port."""
    threading.Thread.__init__(self)

    self.host, self.port = host, port
    self.sock = None
    self.queue = Queue()
    self.maxQueueSize = maxQueueSize
    self.daemon = True


  def run(self):
    """Run the thread."""
    while True:
      try:
        try:
          name, value, valueType, stamp = self.queue.get()
        except TypeError:
          break
        self.log(name, value, valueType, stamp)
      finally:
        self.queue.task_done()


  def connect(self):
    """Connects to the Graphite server if not already connected."""
    if self.sock is not None:
      return
    backoff = 0.01
    while True:
      try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)
        sock.connect((self.host, self.port))
        self.sock = sock
        return
      except socket.error:
        time.sleep(random.uniform(0, 2.0*backoff))
        backoff = min(backoff*2.0, 5.0)


  def disconnect(self):
    """Disconnect from the Graphite server if connected."""
    if self.sock is not None:
      try:
        self.sock.close()
      except socket.error:
        pass
      finally:
        self.sock = None


  def _sendMsg(self, msg):
    """Send a line to graphite. Retry with exponential backoff."""
    if not self.sock:
      self.connect()
    backoff = 0.001
    while True:
      try:
        self.sock.sendall(msg)
        break
      except socket.error:
        logging.warning('Graphite connection error', exc_info = True)
        self.disconnect()
        time.sleep(random.uniform(0, 2.0*backoff))
        backoff = min(backoff*2.0, 5.0)
        self.connect()


  def _sanitizeName(self, name):
    """Sanitize a metric name."""
    return name.replace(' ', '-')


  def log(self, name, value, valueType=None, stamp=None):
    """Log a named numeric value. The value type may be 'value',
    'count', or None."""
    if type(value) == float:
      form = "%s%s %2.2f %d\n"
    else:
      form = "%s%s %s %d\n"

    if valueType is not None and len(valueType) > 0 and valueType[0] != '.':
      valueType = '.' + valueType

    if not stamp:
      stamp = time.time()

    self._sendMsg(form % (self._sanitizeName(name), valueType or '', value, stamp))


  def enqueue(self, name, value, valueType=None, stamp=None):
    """Enqueue a call to log."""
    # If queue is too large, refuse to log.
    if self.maxQueueSize and self.queue.qsize() > self.maxQueueSize:
      return
    # Stick arguments into the queue
    self.queue.put((name, value, valueType, stamp))


  def flush(self):
    """Block until all stats have been sent to Graphite."""
    self.queue.join()


  def shutdown(self):
    """Shut down the background thread."""
    self.queue.put(None)
    self.flush()



class AtomicValue(object):
  """Stores a value, atomically."""

  def __init__(self, val):
    self.lock = threading.RLock()
    self.value = val


  def update(self, function):
    """Atomically apply function to the value, and return the old and new values."""
    with self.lock:
      oldValue = self.value
      self.value = function(oldValue)
      return oldValue, self.value


  def getAndSet(self, newVal):
    """Sets a new value while returning the old value"""
    return self.update(lambda _: newVal)[0]


  def addAndGet(self, val):
    """Adds val to the value and returns the result"""
    return self.update(lambda x: x + val)[1]



class EWMA(object):
  """
  An exponentially-weighted moving average.

  Ported from Yammer metrics.
  """

  M1_ALPHA = 1 - exp(-5 / 60.0)
  M5_ALPHA = 1 - exp(-5 / 60.0 / 5)
  M15_ALPHA = 1 - exp(-5 / 60.0 / 15)

  TICK_RATE = 5 # Every 5 seconds


  @classmethod
  def oneMinute(cls):
    """Creates an EWMA configured for a 1 min decay with a 5s tick"""
    return EWMA(cls.M1_ALPHA, 5)


  @classmethod
  def fiveMinute(cls):
    """Creates an EWMA configured for a 5 min decay with a 5s tick"""
    return EWMA(cls.M5_ALPHA, 5)


  @classmethod
  def fifteenMinute(cls):
    """Creates an EWMA configured for a 15 min decay with a 5s tick"""
    return EWMA(cls.M15_ALPHA, 5)


  def __init__(self, alpha, interval):
    self.alpha = alpha
    self.interval = interval
    self.rate = 0
    self._uncounted = AtomicValue(0)
    self._initialized = False


  def update(self, val):
    """Adds this value to the count to be averaged"""
    self._uncounted.addAndGet(val)


  def tick(self):
    """Updates rates and decays"""
    count = self._uncounted.getAndSet(0)
    instantRate = float(count) / self.interval

    if self._initialized:
      self.rate += (self.alpha * (instantRate - self.rate))
    else:
      self.rate = instantRate
      self._initialized = True




from six import StringIO

from greplin import scales
from greplin.scales import formats, util

from bottle import abort, request, response, run, Bottle
import functools

def bottlestats(server_name, path=''):
    """Renders a GET request, by showing this nodes stats and children."""
    path = path.lstrip('/')
    parts = path.split('/')
    if not parts[0]:
        parts = parts[1:]
    stat_dict = util.lookup(scales.getStats(), parts)

    if stat_dict is None:
        abort(404, "Not Found")
        return

    output = StringIO()
    output_format = request.query.get('format', 'html')
    query = request.query.get('query', None)
    if output_format == 'json':
        response.content_type = "application/json"
        formats.jsonFormat(output, stat_dict, query)
    elif output_format == 'prettyjson':
        formats.jsonFormat(output, stat_dict, query, pretty=True)
        response.content_type = "application/json"
    else:
        formats.htmlHeader(output, '/' + path, server_name, query)
        formats.htmlFormat(output, tuple(parts), stat_dict, query)
        response.content_type = "text/html"

    return output.getvalue()

def register_stats_handler(app, server_name, prefix='/status/'):
    """Register the stats handler with a Flask app, serving routes
    with a given prefix. The prefix defaults to '/_stats/', which is
    generally what you want."""
    if not prefix.endswith('/'):
        prefix += '/'
    handler = functools.partial(bottlestats, server_name)

    app.get(prefix, callback=handler)
    app.get(prefix + '<path:path>', callback=handler)





#!/usr/bin/env python
# Copyright 2011 The greplin-twisted-utils Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script for greplin-twisted-utils."""

try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup

setup(name='scales',
      version='1.0.7',
      description='Stats for Python processes',
      license='Apache',
      author='Greplin, Inc.',
      author_email='opensource@greplin.com',
      url='https://www.github.com/Cue/scales',
      install_requires=[
        'simplejson>=3.0.0',
        'six',
      ],
      package_dir = {'':'src'},
      packages = [
        'greplin',
        'greplin.scales',
      ],
      namespace_packages = [
        'greplin',
      ],
      tests_require = [
          'nose',
      ],
      test_suite = 'nose.collector',
      zip_safe = True,
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
    ],
)



# Copyright 2011 The scales Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a Flask request handler for status reporting."""


from greplin import scales
from greplin.scales import formats, util

from flask import request, abort

from six import StringIO

import functools


def statsHandler(serverName, path=''):
  """Renders a GET request, by showing this nodes stats and children."""
  path = path.lstrip('/')
  parts = path.split('/')
  if not parts[0]:
    parts = parts[1:]
  statDict = util.lookup(scales.getStats(), parts)

  if statDict is None:
    abort(404, 'No stats found with path /%s' % '/'.join(parts))

  output = StringIO()
  outputFormat = request.args.get('format', 'html')
  query = request.args.get('query', None)
  if outputFormat == 'json':
    formats.jsonFormat(output, statDict, query)
  elif outputFormat == 'prettyjson':
    formats.jsonFormat(output, statDict, query, pretty=True)
  else:
    formats.htmlHeader(output, '/' + path, serverName, query)
    formats.htmlFormat(output, tuple(parts), statDict, query)

  return output.getvalue()


def registerStatsHandler(app, serverName, prefix='/status/'):
  """Register the stats handler with a Flask app, serving routes
  with a given prefix. The prefix defaults to '/status/', which is
  generally what you want."""
  if prefix[-1] != '/':
    prefix += '/'
  handler = functools.partial(statsHandler, serverName)
  app.add_url_rule(prefix, 'statsHandler', handler, methods=['GET'])
  app.add_url_rule(prefix + '<path:path>', 'statsHandler', handler, methods=['GET'])


def serveInBackground(port, serverName, prefix='/status/'):
  """Convenience function: spawn a background server thread that will
  serve HTTP requests to get the status. Returns the thread."""
  import flask, threading
  from wsgiref.simple_server import make_server
  app = flask.Flask(__name__)
  registerStatsHandler(app, serverName, prefix)
  server = threading.Thread(target=make_server('', port, app).serve_forever)
  server.daemon = True
  server.start()
  return server
